{************************************************}
{                                                }
{   Turbo Pascal for Windows: Paint Demo         }
{   Bitmaps unit                                 }
{   Copyright (c) 1991 by Borland International  }
{                                                }
{************************************************}

unit Bitmaps;

{
  This unit augments the HBitmap type by implementing load and store of the
  bitmap to a file.

  Presently operates only on Windows format for bitmaps.
}
{$S-,R-}

interface

uses WinTypes, WinProcs;

{
  Read a bitmap from file (full pathname).
  Returns 0 if error or HBitmap.
}
function LoadBitmapFile(FileName: PChar): HBitmap;

{
  Write a bitmap to file (full pathname).
  Returns 0 if error else non-zero.
}
function StoreBitmapFile(FileName: PChar; HBM: HBitmap): Integer;

{************************************************}
implementation

procedure AHIncr; far; external 'KERNEL' index 114;

const
  MaxIO: Word = 65534;	{ Max number bytes handled in single IO operation }
  OneIO: Word = 32768;  { Number bytes handled per huge IO operation }
  BMType: Word = ord('M') shl 8 + ord('B'); { 'BM' = $4D42 }

type
  { Implement long pointers with two adjacent words }
  PtrRec = record Lo, Hi: Word end;
  IOFunction = function(FP: integer; Buf: PChar; Size: Integer): Word;

{
  Perform IO operation in chunks to avoid memory segment crossings.
  Returns 0 if error else non-zero.
}
function HugeIO(IOFunc: IOFunction; F: Integer; P: Pointer; Size: Longint)
               : Word;
var
  L, N: Longint;	       	 { L maintains total bytes }
begin   			 { N maintains bytes for current pass }
  HugeIO := 1;
  L := 0;
  while L < Size do
  begin
    N := Size - L;
    if N > OneIO then N := OneIO;
    if IOFunc(F,
              { 
                Compute the segment and offset reached.
                The Hi word of P contains the initial segment.
                Think of the following as performing arithmetic
                  modulo segment-size, since the size of a segment
                  fills one word:
                The Hi word of L contains the number of segments crossed
                  (the size of one segment fills the Lo word, so Hi word
                  will roll over as segments are filled).
                  Multiply by Ofs(AHIncr) to obtain the number used to
                  indicate this number of segments.
                The Lo word of L contains the number of bytes already
                  passed in the present segment.
               }
	       Ptr(PtrRec(P).Hi + PtrRec(L).Hi * Ofs(AHIncr),
	           PtrRec(L).Lo),
	       Integer(N))  	{ Guaranteed to be in Integer range }
       <> N then
    begin HugeIO := 0; exit end; { abnormal termination }
    Inc(L, N);
  end;
end;

{
  Read a bitmap from file (full pathname).
  Returns 0 if error or HBitmap.
}
function LoadBitmapFile(FileName: PChar): HBitmap;
label
  Exit1, Exit2, Exit3;
var
  F: Integer;			{ File }
  H: THandle;			{ Handle to memory for bitmap }
  DC: HDC;			{ Drawing context for application }
  Size, N: Longint;		{ Size of bitmap, Size of color spec }
  P: PBitmapInfo;		{ Windows bitmap format info header }
  Header: TBitmapFileHeader;    { Bitmap file header }

begin
  LoadBitmapFile := 0;
  F := _LOpen(FileName, of_Read);
  if F = -1 then goto Exit1;

  { read in the Bitmap file header }
  if (_LRead(F, @Header, SizeOf(Header)) <> SizeOf(Header)) or
    (Header.bfType <> BMType) then goto Exit2;

  { read the rest of the file }
  Size := Header.bfSize - SizeOf(TBitmapFileHeader);
  H := GlobalAlloc(gmem_Moveable, Size);	{ Allocate the memory }
  if H = 0 then goto Exit2;
  P := GlobalLock(H);				{ Lock it down }
  if HugeIO(_LRead, F, P, Size) = 0 then goto Exit3;

  { Make sure it's Windows format }
  if P^.bmiHeader.biSize <> SizeOf(TBitmapInfoHeader) then goto Exit3;

  { Compute the size of the color description that follows the info header }
  N := P^.bmiHeader.biBitCount;
  if N = 24 then N := 0 else N := (1 shl N) * SizeOf(TRGBQuad);
  
  { actually create the Bitmap }
  DC := GetDC(0);
  LoadBitmapFile := CreateDIBitmap(DC, P^.bmiHeader,
    cbm_Init, Ptr(PtrRec(P).Hi, SizeOf(TBitmapInfoHeader) + N),
    P^, dib_RGB_Colors);

  { clean up }
  ReleaseDC(0, DC);

  Exit3:
  GlobalUnlock(H);
  GlobalFree(H);

  Exit2:
  _LClose(F);

  Exit1:
end;


{
  Write a bitmap to file (full pathname).
  Returns 0 if error else non-zero.
}
function StoreBitmapFile(FileName: PChar; HBM: HBitmap): Integer;
  label
    Exit0, Exit1, Exit2, Exit3;
  var
    BM:   TBitmap;		{ Bitmap information }
    BFH:  TBitmapFileHeader;	{ Bitmap file information }
    BIP:  PBitmapInfo;		{ Part of bitmap file information }
    DC:   HDC;			{ Drawing context }

    HMem: THandle;		{ Handle to memory for bitmap }
    Buf:  Pointer;		{ Memory for bitmap }

    ColorSize, DataSize: Longint; { Size needed to store Color/Data }
    BitCount: Word;		{ Number of bits per pixel }
    FP: Integer;		{ File }

  {
    Takes the size in bits and returns the (aligned) size in bytes.
    Bitmap data format requires word alignment.
  }
  function bmAlignDouble(Size: Longint): Longint;
  begin
    bmAlignDouble := (Size + 31) div 32 * 4;
  end;

begin
   StoreBitmapFile := 0;
   FP := _lcreat(FileName, 0);
   if FP = -1 then goto Exit0;

   { Get the information about the Bitmap }
   if GetObject(HBM, sizeOf(TBitmap), @BM) = 0 then goto Exit1;

   BitCount := bm.bmPlanes * bm.bmBitsPixel;
   if (BitCount <> 24) then
     ColorSize := sizeOf(TRGBQuad) * (1 shl BitCount - 1)
   else
     ColorSize := 0;
   DataSize := bmAlignDouble(bm.bmWidth * BitCount) * bm.bmHeight;

   GetMem(BIP, sizeOf(TBitmapInfoHeader) + ColorSize);
   if BIP = nil then goto Exit1;

   { Fill in the Bitmap info header }
   with BIP^.bmiHeader do
   begin
     biSize := sizeOf(TBitmapInfoHeader);
     biWidth := bm.bmWidth;
     biHeight := bm.bmHeight;
     biPlanes := 1;
     biBitCount := BitCount;
     biCompression := 0;
     biSizeImage := DataSize;
     biXPelsPerMeter := 0;
     biYPelsPerMeter := 0;
     biClrUsed := 0;
     biClrImportant := 0;
   end;

   { Fill in the file header }
   with BFH do
   begin
     bfOffBits := sizeOf(BFH) + sizeOf(TBitmapInfo) + ColorSize;
     bfReserved1 := 0;
     bfReserved2 := 0;
     bfSize :=  bfOffBits + DataSize;
     bfType := BMType;
   end;

   { Create the memory Bitmap }
   { GetMem can only allocate up to 64K }
   HMem := GlobalAlloc(gmem_Fixed, DataSize); 
   if HMem = 0 then goto Exit2;
   Buf := GlobalLock(HMem);

   { Get the bitmap bits in device independent format }
   DC := GetDC(0);
   if GetDIBits(DC, hbm, 0, DataSize, Buf, BIP^, dib_RGB_Colors) = 0 then
     goto Exit3;
   ReleaseDC(0, DC);
     
   { Write to file }
   _lwrite(FP, @BFH, sizeOf(BFH));
   _lwrite(FP, PChar(BIP), sizeOf(TBitmapInfo) + ColorSize);
   HugeIO(_lwrite, FP, Buf, DataSize);

   StoreBitmapFile := 1;

   { Clean up }
   Exit3:
   GlobalUnlock(HMem);
   GlobalFree(HMem);

   Exit2:
   FreeMem(BIP, sizeOf(TBitmapInfoHeader));

   Exit1:
   _lclose(FP);

   Exit0:
end;

end.
