{************************************************}
{                                                }
{   Turbo Pascal for Windows: Paint Demo         }
{   Canvas unit                                  }
{   Copyright (c) 1991 by Borland International  }
{                                                }
{************************************************}

unit Canvas;

{
  This unit supplies the drawing canvas for the paint program, that is, the
  window where drawing actually takes place.

  The Canvas is responisble for maintaining the screen state, including
  updating the cursor, directing input to the currently selected drawing
  tool and managing the enabling of certain menu items (cut/copy/paste/etc).
}

interface

uses PaintDef, Bitmaps,
     WinTypes, WinProcs, WObjects;

type

  PCanvas = ^TCanvas;
  TCanvas = object(TWindow)
    State: PState;

    Drawing: Boolean;		{ In the process of drawing }
    OverSelection: Boolean;	{ Cursor is the 'over selection' cursor }

    { Creation and destruction }
    constructor Init(AParent: PWindowsObject; AState: PState);
    destructor Done; virtual;

    { Display }
    procedure Paint(PaintDC: HDC; var PaintInfo: TPaintStruct); virtual;
    procedure PaintSelection(DC: HDC);
    procedure ScreenSave;

    { Menu initiated actions }
    { File }
    function Load(FileName: PChar): Integer;
    function Store(FileName: PChar): Integer;

    { Edit }
    procedure CopyToClipBoard(DC: HDC; Left, Top, Width, Height: Integer);
    procedure Erase(Left, Top, Width, Height: Integer);
    procedure PickUpSelection(aDC: HDC; Left, Top, Width, Height: Integer);
    procedure ReleaseSelection;

    procedure Cut;
    procedure Copy;
    procedure Paste;
    procedure Delete;
    procedure ClearAll;

    { Window manager responses }
    { Mouse initiated actions }
    procedure WMLButtonDown(var Msg: TMessage);
      virtual wm_First + wm_LButtonDown;
    procedure WMLButtonUp(var Msg: TMessage);
      virtual wm_First + wm_LButtonUp;
    procedure WMMouseMove(var Msg: TMessage);
      virtual wm_First + wm_MouseMove;
    procedure wmSetCursor(var Msg: TMessage);
      virtual wm_First + wm_SetCursor;

  end;


{************************************************}
implementation

const

  BitmapWidth  = 640;			{ Bitmaps are currently fixed size }
  BitmapHeight = 480;

{ 
  Create a new canvas and initialize the selection.
}
constructor TCanvas.Init(AParent: PWindowsObject; AState: PState);
var
  DC, MemDC: HDC;
begin
  TWindow.Init(AParent, nil);
  Attr.Style := ws_Border or ws_Child or ws_Visible;
  State := AState;
  
  { Initialize the selection }
  SetRectEmpty(State^.Selection);
  State^.SelectionBM := 0;

  Drawing := False;
  OverSelection := False;

  { Create the off-screen bitmap to record the image drawn }
  DC := GetDC(0);
  MemDC := CreateCompatibleDC(DC);
  State^.Bitmap := CreateCompatibleBitmap(DC, BitmapWidth, BitmapHeight);
  ReleaseDC(0, DC);
  
  { Paint it white }
  State^.Bitmap := SelectObject(MemDC, State^.Bitmap);
  PatBlt(MemDC, 0, 0, BitmapWidth, BitmapHeight, whiteness);
  State^.Bitmap := SelectObject(MemDC, State^.Bitmap);

  { Clean up }
  DeleteDC(MemDC);
end;

{
  Destroy the off-screen bitmap before dying.
}
destructor TCanvas.Done;
begin
  DeleteObject(State^.Bitmap);
  TWindow.Done;
end;

{ Display }
{
  Update the screen display from the off-screen bitmap. Highlight the 
  selection if there is one.
}
procedure TCanvas.Paint(PaintDC: HDC; var PaintInfo: TPaintStruct);
var
  MemDC: HDC;			{ For the off-screen bitmap }
  R: TRect;			{ The window client area }
begin
  { Set up the drawing contexts }
  MemDC := CreateCompatibleDC(PaintDC);
  State^.Bitmap := SelectObject(MemDC, State^.Bitmap);

  { Copy from the off-screen bitmap to the screen }
  GetClientRect(HWindow, R);
  BitBlt(PaintDC, 0, 0, R.right, R.bottom, MemDC, 0, 0, srcCopy);

  { Highlight the selection }  
  PaintSelection(PaintDC);

  { Clean up }
  State^.Bitmap := SelectObject(MemDC, State^.Bitmap);
  DeleteDC(MemDC);
end;

{
  Highlight the selection, if there is one, by drawing a dotted line around
  it. If there is a selection bitmap display it.
}
procedure TCanvas.PaintSelection(DC: HDC);
var SelDC: HDC;			{ For the selection bitmap }
begin
  if not IsRectEmpty(State^.Selection) then
    begin

      { Draw the selecton bitmap }
      if State^.SelectionBM <> 0 then
        begin

          { Set up the drawing context }
          SelDC := CreateCompatibleDC(DC);
	  State^.SelectionBM := SelectObject(SelDC, State^.SelectionBM);

          { Copy the bits to the screen }
	  with State^.Selection do
	    BitBlt(DC, Left, Top, Right-Left, Bottom-Top, SelDC, 0, 0, SrcCopy);

	  { Clean up }
	  State^.SelectionBM := SelectObject(SelDC, State^.SelectionBM);
	  DeleteDC(SelDC);
	end;
      
      { Draw a dotted line marking the selected area }
      SetROP2(DC, r2_CopyPen);
      SelectObject(DC, DashedPen);
      SelectObject(DC, GetStockObject(Null_Brush));
      with State^.Selection do
	Rectangle(DC, Left, Top, Right, Bottom);
    end;
end;

{
  Save the image on the screen to the off-screen bitmap.
}
procedure TCanvas.ScreenSave;
var
  DC, MemDC: HDC;		{ For the screen, off-screen bitmaps }
  R: TRect;			{ The window client area }
begin
  { Set up the drawing contexts }
  DC := GetDC(HWindow);
  MemDC := CreateCompatibleDC(DC);
  State^.Bitmap := SelectObject(MemDC, State^.Bitmap);

  { Copy the bits }
  GetClientRect(HWindow, R);
  BitBlt(MemDC, 0, 0, R.right, R.bottom, DC, 0, 0, srcCopy);
  State^.Bitmap := SelectObject(MemDC, State^.Bitmap);

  { Clean up }
  DeleteDC(MemDC);
  ReleaseDC(HWindow, DC);
end;  


{ Menu initiated functions }
{ File }

{
  Read a bitmap from a file into the current drawing canvas. Returns 0 on 
  error, otherwise non-zero.
}
function TCanvas.Load(FileName: PChar): Integer;

  function Smaller(A, B: Integer): Integer;
  begin
    if A < B then Smaller := A else Smaller := B;
  end;

var
  HBM: HBitmap;			{ The new bitmap }
  BM: TBitmap;			{ Information about the new bitmap }
  DC, NewDC, MemDC: HDC;	{ For screen, new and off-screen bitmaps }
  R: TRect;			{ The window client area }
begin
    Load := 1;

    { Actually read in the bitmap }
    HBM := LoadBitmapFile(FileName);
    if HBM = 0 then		{ Failure }
    begin
      Load := 0;
      exit;
    end;

    { Mark the bitmap as unmodified }
    State^.IsDirtyBitmap := False;

    { Copy the bitmap onto the (fixed size) screen and memory bitmaps }
    GetObject(HBM, sizeOf(BM), @BM);	{ Information about the new bitmap }

    { Set up the drawing contexts }
    DC := GetDC(HWindow);
    NewDC := CreateCompatibleDC(DC);
    MemDC := CreateCompatibleDC(DC);
    HBM := SelectObject(NewDC, HBM);
    State^.Bitmap := SelectObject(MemDC, State^.Bitmap);

    { White out old image so it doesn't show anywhere }
    PatBlt(MemDC, 0, 0, BitmapWidth, BitmapHeight, whiteness);

    { Copy in the new image to both screen and off-screen bitmaps }
    BitBlt(
      MemDC, 0, 0,
      Smaller(bm.bmWidth, BitmapWidth), Smaller(bm.bmHeight, BitmapHeight),
      NewDC, 0, 0, srcCopy);
    GetClientRect(HWindow, R);
    BitBlt(DC, 0, 0, R.right, R.bottom, MemDC, 0, 0, srcCopy);

    { Clean up }
    HBM := SelectObject(NewDC, HBM);
    State^.Bitmap := SelectObject(MemDC, State^.Bitmap);
    DeleteDC(NewDC);
    DeleteDC(MemDC);
    ReleaseDC(HWindow, DC);
    DeleteObject(HBM);

    SetRectEmpty(State^.Selection);
end;

{
  Write the current image out to a file. Returns 0 if error, otherwise
  non-zero.
}
function TCanvas.Store(FileName: PChar): Integer;
var I: Integer;			{ Result from the actual write }
begin
  I := StoreBitmapFile(FileName, State^.Bitmap);
  State^.IsDirtyBitmap := I <> 1; { Mark the bitmap unmodified if successful }
  Store := I;
end;

{ Edit }

{
  Copy the indicated bits of bitmap in the drawing context to the clipboard.
  Copying to the clipboard is done by transferring a bitmap to the clipboard.
  Once the clipboard has been passed this bitmap, it is no longer owned by
  the application, so a new bitmap is created expressly for this purpose.
}
procedure TCanvas.CopyToClipBoard(DC: HDC; Left, Top, Width, Height: Integer);
var
  CopyDC: HDC;			{ For the new bitmap }
  CopyBitmap: HBitmap;		{ The new bitmap }
begin
  { Make sure clipboard is available and can be copied to }
  if OpenClipBoard(HWindow) and EmptyClipBoard then
    begin

      { Create the new bitmap }
      CopyDC := CreateCompatibleDC(DC);
      CopyBitmap := CreateCompatibleBitmap(DC, Width, Height);
      CopyBitmap := SelectObject(CopyDC, CopyBitmap);
      BitBlt(CopyDC, 0, 0, Width, Height, DC, Left, Top, SrcCopy);
      CopyBitmap := SelectObject(CopyDC, CopyBitmap);
      
      { Transfer the new bitmap to the clipboard }
      SetClipBoardData(cf_Bitmap, CopyBitmap);

      { Clean up }
      CloseClipBoard;
      DeleteDC(CopyDC);
    end;
end;

{
  White out the rectangle indicated on the off-screen bitmap.
}
procedure TCanvas.Erase(Left, Top, Width, Height: Integer);
var
  MemDC: HDC;			{ For the off-screen bitmap }
begin
  
  { Set up the drawing context }
  MemDC := CreateCompatibleDCW(HWindow);
  State^.Bitmap := SelectObject(MemDC, State^.Bitmap);
  
  { White out the rectangle }
  PatBlt(MemDC, Left, Top, Width, Height, Whiteness);
  State^.Bitmap := SelectObject(MemDC, State^.Bitmap);

  { Clean up }
  DeleteDC(MemDC);
end;

{
  Make the current selection into a selection bitmap. Note that this should
  (and can be) only invoked when the SelectTool is active. (Otherwise there
  could be no selection.
}
procedure TCanvas.PickUpSelection(aDC: HDC; Left, Top, Width, Height: Integer);
begin
  State^.PaintTool^.PickUpSelection(aDC, Left, Top, Width, Height);
end;

{
  Release the current selection without saving the bits. Also gray out the
  appropriate menu items.
}
procedure TCanvas.ReleaseSelection;
begin
  State^.PaintTool^.ReleaseSelection;
  EnableMenuItem(GetMenu(Parent^.HWindow), cm_EditCut, mf_grayed);
  EnableMenuItem(GetMenu(Parent^.HWindow), cm_EditCopy, mf_grayed);
  EnableMenuItem(GetMenu(Parent^.HWindow), cm_EditDelete, mf_grayed);
end;

{
  Copy the current selection to the clipboard and white out the hole.
}
procedure TCanvas.Cut;
begin
  Copy;
  Delete;
end;

{
  Copy the current selection to the clipboard.
}
procedure TCanvas.Copy;
var
  MemDC: HDC;			{ For the bitmap to copy from }
begin
  MemDC := CreateCompatibleDCW(HWindow);
  if State^.SelectionBM <> 0 then
    
    { Use the selection bitmap }
    begin
      State^.SelectionBM := SelectObject(MemDC, State^.SelectionBM);
      with State^.Selection do
        CopyToClipBoard(MemDC, 0, 0, Right-Left, Bottom-Top);
      State^.SelectionBM := SelectObject(MemDC, State^.SelectionBM);
    end
  else
    
    { Copy from the off-screen bitmap }
    begin
      State^.Bitmap := SelectObject(MemDC, State^.Bitmap);
      with State^.Selection do
        CopyToClipBoard(MemDC, Left, Top, Right-Left, Bottom-Top);
      State^.Bitmap := SelectObject(MemDC, State^.Bitmap);
    end;
  DeleteDC(MemDC);
end;

{
  Retrieve what is in the clipboard and make it the current selection bitmap.
  The clipboard retains ownership of the retrieved bitmap, so it must be
  copied into a new selection bitmap.
}
procedure TCanvas.Paste;
var
  DC, ClipDC: HDC;		{ For screen and clipboard bitmaps }
  ClipBitmap: HBitmap;		{ The clipboard bitmap }
  BM: TBitmap;			{ Information on the clipboard bitmap }
begin

  { Make sure the clipboard is available }
  if OpenClipBoard(HWindow) then
    begin

      { Set up the drawing contexts }
      DC := GetDC(HWindow);
      ClipDC := CreateCompatibleDC(DC);

      { Retrieve the clipboard bitmap }
      ClipBitmap := GetClipBoardData(cf_Bitmap);
      CloseClipBoard;

      { Make sure the retrieve succeeded and make it the selection bitmap }
      if (ClipBitmap <> 0) and
         { Get information about the bitmap }
	 (GetObject(ClipBitmap, sizeOf(TBitmap), @BM) <> 0) then
        begin
	  ClipBitmap := SelectObject(ClipDC, ClipBitmap);
	  PickUpSelection(ClipDC, 0, 0, bm.bmWidth, bm.bmHeight);
	  ClipBitmap := SelectObject(ClipDC, ClipBitmap);
	  PaintSelection(DC);
	end;

      { Clean up }
      DeleteDC(ClipDC);
      ReleaseDC(HWindow, DC);
    end;
end;

{
  White out the selected area or release the selection bitmap.
}
procedure TCanvas.Delete;
begin
  if State^.SelectionBM = 0 then
    with State^.Selection do Erase(Left, Top, Right-Left, Bottom-Top);
  ReleaseSelection;
end;


{ 
  White out the entire canvas.
}
procedure TCanvas.ClearAll;
var
  DC, MemDC: HDC;		{ For the screen and off-screen bitmaps }
  R: TRect;			{ The window client area }
begin
  GetClientRect(HWindow, R);
  InvalidateRect(HWindow, @R, False);
  ReleaseSelection;
  Erase(0, 0, BitmapWidth, BitmapHeight);
end;

{ Window manager responses }
{ Mouse initiated actions }

{
  Start the selected drawing tool drawing.
}
procedure TCanvas.WMLButtonDown(var Msg: TMessage);
begin
  if not Drawing then
  begin
    { Let subsequent Mouse Moves and Mouse Ups know that drawing is in 
      progress, i.e., that the initial mouse down occurred in the right
      window.
    }
    Drawing := True;
 
    { Mark the bitmap as modified }
    State^.IsDirtyBitmap := True;

    { Tell the current tool to start drawing }
    State^.PaintTool^.MouseDown(HWindow, Integer(Msg.LParamLo),
					 Integer(Msg.LParamHi), State);
  end;
end;

{
  If drawing is in progress, tell the currently selected tool about the
  Mouse Move.
}
procedure TCanvas.WMMouseMove(var Msg: TMessage);
begin
  if Drawing then
    State^.PaintTool^.MouseMove(Integer(Msg.LParamLo),
				Integer(Msg.LParamHi))
end;

{
  If drawing is in progress, record the altered state of the image by either
  copying the screen bitmap to the off-screen bitmap or high-lighting the
  new selection. Tell the currently selected tool that the mouse is up.
  Enable/disable menus appropriately.
}
procedure TCanvas.WMLButtonUp(var Msg: TMessage);
var
  DC: HDC;				{ For the screen bitmap }
begin
  if Drawing then
  begin
    State^.PaintTool^.MouseUp;
    Drawing := False;
    if IsRectEmpty(State^.Selection) then
      begin
        ScreenSave;
	EnableMenuItem(GetMenu(Parent^.HWindow), cm_EditCut, mf_grayed);
	EnableMenuItem(GetMenu(Parent^.HWindow), cm_EditCopy, mf_grayed);
	EnableMenuItem(GetMenu(Parent^.HWindow), cm_EditDelete, mf_grayed);
      end
    else
      begin
        DC := GetDC(HWindow);
	PaintSelection(DC);
        ReleaseDC(HWindow, DC);
	EnableMenuItem(GetMenu(Parent^.HWindow), cm_EditCut, mf_enabled);
	EnableMenuItem(GetMenu(Parent^.HWindow), cm_EditCopy, mf_enabled);
	EnableMenuItem(GetMenu(Parent^.HWindow), cm_EditDelete, mf_enabled);
      end;
  end;
end;

{
  When the cursor is over the canvas, change the cursor to the cursor
  associated with the selected tool. If the cursor is over the selection
  use the standard arrow cursor.
}
procedure TCanvas.WMSetCursor(var Msg: TMessage);
var Pt: TPoint;
begin
  GetCursorPos(Pt);		{ In global coordinates }
  ScreenToClient(HWindow, Pt);  { In window client local coordinates }
  if PtInRect(State^.Selection, Pt) then
    SetCursor(LoadCursor(0, idc_Arrow))
  else
    SetCursor(State^.PaintTool^.Cursor)
end;

end.